//Jose M. Pea
//CompBio, IFM, LiU
//24/08/2006

//Approximate counting of CPDAGs via MCMC.

///////////////////////////////////includes

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

///////////////////////////////////global variables

int n_vars, n_its;

///////////////////////////////////sink

int sink(int x, int **cpdag, int *marked)

{int i;

for(i=0;i<n_vars;i++)
	if(marked[i]==0 && cpdag[x][i]==1)
		return 0;

return 1;
}

///////////////////////////////////complete

int complete(int x, int **cpdag, int *marked)

{int i, j;

for(i=0;i<n_vars;i++)
	if(marked[i]==0 && cpdag[i][x]==2)
		for(j=0;j<n_vars;j++)
			if(j!=i && marked[j]==0 && cpdag[j][x]!=0 && cpdag[i][j]==0 && cpdag[j][i]==0)
				return 0;

return 1;
}

///////////////////////////////////cpdag_2_dag 

int cpdag_2_dag(int **cpdag, int **dag)

//(Dor and Tarsi 1992).
//cpdag must be a PDAG according to Dor and Tarsi.
//I think that is not really necessary, i.e. it may contain directed cycles.

{int i, j, k, stop, stop2, *marked;

marked=new int[n_vars];
for(i=0;i<n_vars;i++)
	{marked[i]=0;
	for(j=0;j<n_vars;j++)
		dag[i][j]=0;
	}

stop=n_vars;
while(stop>0)
	{stop2=stop;
	for(j=0;j<n_vars;j++)
		if(marked[j]==0 && sink(j,cpdag,marked)==1 && complete(j,cpdag,marked)==1)
			{stop--;
			marked[j]=1;

			for(k=0;k<n_vars;k++)
				if(marked[k]==0 && cpdag[k][j]!=0)
					dag[k][j]=1; 
			}

	if(stop2==stop)
		{delete [] marked;	
		
		return 0;
		}
	}

delete [] marked;

return 1;
}

///////////////////////////////////node_levels

int node_levels(int **dag, int *level)

//dag may contain directed cyles.
//In that case, -1 is returned.

{int i, j, max_levels, stop;

for(i=0;i<n_vars;i++)
	level[i]=0;

max_levels=0;
stop=0;
while(stop==0)
	{stop=1;
	for(i=0;i<n_vars;i++)
		if(level[i]==max_levels)
			for(j=0;j<n_vars;j++)
				 if(dag[i][j]==1)
					{if(max_levels<n_vars-1)
						{stop=0;
						level[j]=max_levels+1;
						}
					else
						return -1;
					}

	 max_levels++;
	}

return max_levels-1;
}

///////////////////////////////////dag_2_cpdag

void dag_2_cpdag(int **dag, int **cpdag)

//(Chickering 1995).

{int i, j, k, aux, aux2, x, y, max_levels, n_arrows, *level;

level=new int[n_vars];

//Order nodes.

max_levels=node_levels(dag,level);

//Order arrows.

for(i=0;i<n_vars;i++)
	for(j=0;j<n_vars;j++)
		cpdag[i][j]=0;

n_arrows=0;
for(k=0;k<=max_levels;k++)
	for(i=0;i<n_vars;i++)
		if(level[i]==k)
			{aux=0;
			while(aux>-1)
				{aux=-1;
			    for(j=0;j<n_vars;j++)
					 if(dag[j][i]==1 && cpdag[j][i]==0)
						{if(aux==-1)
							 aux=j;
						 else
							 if(level[j]>level[aux])
								aux=j;
						}

				 if(aux>-1)
					{n_arrows++;
					cpdag[aux][i]=n_arrows; 
					}		
				}
			 }

//Find essential arrows.

aux2=n_arrows+1;
while(n_arrows>0)
	{aux=aux2;
	for(i=0;i<n_vars;i++)
		for(j=0;j<n_vars;j++)
			if(cpdag[i][j]>0 && cpdag[i][j]<aux)
				{x=i;
				y=j;
				aux=cpdag[i][j];
				}

	 aux=0;
	 for(i=0;i<n_vars;i++)
		if(cpdag[i][x]==-1)
			{if(cpdag[i][y]==0) 
				{for(j=0;j<n_vars;j++)
					if(cpdag[j][y]>0)
						{cpdag[j][y]=-1;
					     n_arrows--;
						}

				i=n_vars;
				aux=1;
				}
			else
				{cpdag[i][y]=-1;
				n_arrows--;
				}
			}

	 if(aux==0)
		{for(i=0;i<n_vars;i++)
			if(cpdag[i][y]!=0 && i!=x && cpdag[i][x]==0)
				{aux=1;
				i=n_vars;
				}

		if(aux==1)
			{for(i=0;i<n_vars;i++)
				if(cpdag[i][y]>0)
					{cpdag[i][y]=-1;
					n_arrows--;
					}
			}
		else
			{for(i=0;i<n_vars;i++)
				if(cpdag[i][y]>0)
					{cpdag[i][y]=-2;
					n_arrows--;
					}
			}
		}
	}

for(i=0;i<n_vars;i++)
	for(j=0;j<n_vars;j++)
		if(cpdag[i][j]==-1)
			cpdag[i][j]=1;
		else
			if(cpdag[i][j]==-2)
				{cpdag[i][j]=2;
				cpdag[j][i]=2;
				}

delete [] level;
}

///////////////////////////////////cpdag_check

int cpdag_check(int **cpdag)

{int i, j, k, l, aux, *level, **dag, **cpdag2;

level=new int[n_vars];
dag=new int*[n_vars];
cpdag2=new int*[n_vars];
for(i=0;i<n_vars;i++)
	{dag[i]=new int[n_vars];
	cpdag2[i]=new int[n_vars];
	}

aux=1;

//No flags (Andersson et al. 1997).
//Some directed cycles are also detected here.

for(i=0;i<n_vars;i++) 
	for(j=0;j<n_vars;j++)
		if(cpdag[i][j]==1)
			for(k=0;k<n_vars;k++)
				if(cpdag[j][k]==2 && cpdag[i][k]!=1)
					{aux=0;
					i=n_vars;
					j=n_vars;
					k=n_vars;
					}

//Chain graph (Andersson et al. 1997).
//I can check it as follows **if cpdag passed the test above**.

if(aux==1)
	if(node_levels(cpdag,level)==-1) 
		aux=0;

//Chordality of chain components (Andersson et al. 1997).

if(aux==1)
	{for(i=0;i<n_vars;i++)
		for(j=0;j<n_vars;j++)
			if(cpdag[i][j]==1)
				cpdag2[i][j]=0;
			else
				cpdag2[i][j]=cpdag[i][j];

	if(cpdag_2_dag(cpdag2,dag)==0)
		aux=0;
	}

//Strongly protected arrows (Andersson et al. 1997).

if(aux==1)
	for(i=0;i<n_vars;i++) 
		for(j=0;j<n_vars;j++)
			if(cpdag[i][j]==1)
				{aux=0;
				for(k=0;k<n_vars;k++)
					if((cpdag[k][i]==1 && cpdag[j][k]==0 && cpdag[k][j]==0) ||
					(k!=i && cpdag[k][j]==1 && cpdag[i][k]==0 && cpdag[k][i]==0) ||
					(cpdag[i][k]==1 && cpdag[k][j]==1))
						{aux=1;
						k=n_vars;
						}

				if(aux==0)
					{for(k=0;k<n_vars;k++)
						for(l=0;l<n_vars;l++)
							if(k!=l && cpdag[i][k]==2 && cpdag[i][l]==2 && cpdag[k][j]==1 && 
							cpdag[l][j]==1 && cpdag[k][l]==0 && cpdag[l][k]==0)
								{aux=1;
								k=n_vars;
								l=n_vars;
								}

					if(aux==0)
						{i=n_vars;
						j=n_vars;
						}
					}
				}
/*
//Existence of consistent extension.

if(aux==1)
	if(cpdag_2_dag(cpdag,dag)==0)
		aux=0;

//No PDAG.

if(aux==1)
	{dag_2_cpdag(dag,cpdag2);

	aux=1;
	for(i=0;i<n_vars-1;i++)
		for(j=i+1;j<n_vars;j++)
			if(cpdag[i][j]!=cpdag2[i][j] || cpdag[j][i]!=cpdag2[j][i])
				{aux=0;
				i=n_vars;
				j=n_vars;
				}	
	}
*/
delete [] level;
for(i=0;i<n_vars;i++)
	{delete [] dag[i];
	delete [] cpdag2[i];
	}
delete [] dag;	
delete [] cpdag2;

return aux;
}

///////////////////////////////////mcmc

int mcmc(double &eq, double &a_l, double &d_l, double &a_a, double &d_a, double &a_v, double &d_v)

//(Andersson et al. 1997).
//MC starting from the empty CPDAG.

{int i, j, x, y, z, **cpdag;

srand((unsigned)time(NULL));

cpdag=new int*[n_vars];
for(i=0;i<n_vars;i++)
	{cpdag[i]=new int[n_vars];
	for(j=0;j<n_vars;j++)
		cpdag[i][j]=0;
	}

eq=(double)0;
a_l=(double)0;
d_l=(double)0;
a_a=(double)0;
d_a=(double)0;
a_v=(double)0;
d_v=(double)0;
for(i=0;i<n_its;i++)
	{switch(rand()%7)
		{case 0:
			eq++;
			break;
		case 1: //Add line.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && cpdag[x][y]==0 && cpdag[y][x]==0)
				{cpdag[x][y]=2;
				cpdag[y][x]=2;
				if(cpdag_check(cpdag)==0)
					{cpdag[x][y]=0;
					cpdag[y][x]=0;						
					}
				else
					a_l++;
				}
			break;
		case 2: //Remove line.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && cpdag[x][y]==2)
				{cpdag[x][y]=0;
				cpdag[y][x]=0;
				if(cpdag_check(cpdag)==0)
					{cpdag[x][y]=2;
					cpdag[y][x]=2;
					}
				else
					d_l++;
				}
			break;
		case 3: //Add arrow.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && cpdag[x][y]==0 && cpdag[y][x]==0)
				{cpdag[x][y]=1;
				if(cpdag_check(cpdag)==0)
					cpdag[x][y]=0;
				else
					a_a++;
				}
			break;
		case 4: //Remove arrow.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && cpdag[x][y]==1)
				{cpdag[x][y]=0;
				if(cpdag_check(cpdag)==0)
					cpdag[x][y]=1;
				else
					d_a++;
				}
			break;
		case 5: //Add immorality.
			x=rand()%n_vars;
			y=rand()%n_vars;
			z=rand()%n_vars;
			if(x!=y && x!=z && y!=z && cpdag[x][y]==0 && cpdag[y][x]==0 && 
			cpdag[x][z]==0 && cpdag[z][x]==0 && cpdag[y][z]==0 && cpdag[z][y]==0)
				{cpdag[x][z]=1;
				cpdag[y][z]=1;
				if(cpdag_check(cpdag)==0)		
					{cpdag[x][z]=0;
					cpdag[y][z]=0;
					}
				else
					a_v++;
				}
			break;
		case 6: //Remove immorality.
			x=rand()%n_vars;
			y=rand()%n_vars;
			z=rand()%n_vars;
			if(x!=y && x!=z && y!=z && cpdag[x][y]==0 && cpdag[y][x]==0 && cpdag[x][z]==1 && cpdag[y][z]==1)
				{cpdag[x][z]=0;
				cpdag[y][z]=0;
				if(cpdag_check(cpdag)==0)		
					{cpdag[x][z]=1;
					cpdag[y][z]=1;
					}	
				else
					d_v++;
				}	
			break;
			}
	}

eq/=(double)n_its;
a_l/=(double)n_its;
d_l/=(double)n_its;
a_a/=(double)n_its;
d_a/=(double)n_its;
a_v/=(double)n_its;
d_v/=(double)n_its;

x=1;
for(i=0;i<n_vars-1;i++)
	for(j=i+1;j<n_vars;j++)
		if(cpdag[i][j]==2)
			{x=0;
			i=n_vars;
			j=n_vars;
			}

for(i=0;i<n_vars;i++)
	delete [] cpdag[i];
delete [] cpdag;

return x;
}

///////////////////////////////////main

void main(int argc, char* argv[])

{int i, n_samples, size1;
double eq, a_l, d_l, a_a, d_a, a_v, d_v, cum_eq, cum_a_l, cum_d_l, cum_a_a, cum_d_a, cum_a_v, cum_d_v;
char name[80];
clock_t clock_init;
FILE *f;

clock_init=clock();

if(argc!=4) 
	{printf_s("\n counting n_vars n_its n_samples \n"); 
	 exit(0);
	}

n_vars=atoi(argv[1]);
n_its=atoi(argv[2]);
n_samples=atoi(argv[3]);

cum_eq=(double)0;
cum_a_l=(double)0;
cum_d_l=(double)0;
cum_a_a=(double)0;
cum_d_a=(double)0;
cum_a_v=(double)0;
cum_d_v=(double)0;
size1=0;
for(i=0;i<n_samples;i++)
	{size1+=mcmc(eq,a_l,d_l,a_a,d_a,a_v,d_v);

	cum_eq+=eq;
	cum_a_l+=a_l;
	cum_d_l+=d_l;
	cum_a_a+=a_a;
	cum_d_a+=d_a;
	cum_a_v+=a_v;
	cum_d_v+=d_v;
	
	if(i%100==99) 
		{printf_s(" %d / %d = %f in %f hours (%f + %f + %f + %f + %f + %f + %f = %f) \n",size1,i+1,
		(double)size1/(double)(i+1),(double)(clock()-clock_init)/(double)(CLOCKS_PER_SEC*3600),
		cum_eq/(double)(i+1),cum_a_l/(double)(i+1),cum_d_l/(double)(i+1),cum_a_a/(double)(i+1),
		cum_d_a/(double)(i+1),cum_a_v/(double)(i+1),cum_d_v/(double)(i+1),
		(cum_eq+cum_a_l+cum_d_l+cum_a_a+cum_d_a+cum_a_v+cum_d_v)/(double)(i+1));
		sprintf_s(name,"counting_%d.txt",n_vars);
		fopen_s(&f,name,"a");
		fprintf_s(f," %d / %d = %f in %f hours (%f + %f + %f + %f + %f + %f + %f = %f) \n",size1,i+1,
		(double)size1/(double)(i+1),(double)(clock()-clock_init)/(double)(CLOCKS_PER_SEC*3600),
		cum_eq/(double)(i+1),cum_a_l/(double)(i+1),cum_d_l/(double)(i+1),cum_a_a/(double)(i+1),
		cum_d_a/(double)(i+1),cum_a_v/(double)(i+1),cum_d_v/(double)(i+1),
		(cum_eq+cum_a_l+cum_d_l+cum_a_a+cum_d_a+cum_a_v+cum_d_v)/(double)(i+1));
		fclose(f);
		}
	}
}
