//Jose M. Pea
//CompBio, IFM, LiU
//29/09/2006

//Approximate counting of LCGs via MCMC.

///////////////////////////////////includes

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

///////////////////////////////////global variables

int n_vars, n_its;

///////////////////////////////////des_build

void des_build(int **lcg, int **des)

{int i, j, k, stop;

for(i=0;i<n_vars;i++)
	{for(j=0;j<n_vars;j++)
		des[i][j]=lcg[i][j];
	des[i][i]=1;
	}

stop=0;
while(stop==0)
	{stop=1;
	for(i=0;i<n_vars;i++)
		for(j=0;j<n_vars;j++)
			if(lcg[i][j]==1)
				for(k=0;k<n_vars;k++)
					if(des[j][k]==1 && des[i][k]==0)
						{des[i][k]=1;
						stop=0;
						}
	}
}

///////////////////////////////////complex_arrows

void complex_arrows(int x, int y, int z, int *visited, int **complex, int **lcg)

{int i, j, stop;

for(i=0;i<n_vars;i++) 
	if(visited[i]==0)
		{stop=0;
		for(j=0;j<n_vars;j++) 
			if(visited[j]==1 && j!=z && (lcg[i][j]==1 || lcg[j][i]==1)) 
				{stop=1;
				j=n_vars;
				}

		if(stop==0)
			{if(lcg[i][z]==1 && lcg[z][i]==0) 
				{complex[x][y]=1;
				complex[i][z]=1;
				}
			else 
				if(lcg[i][z]==1 && lcg[z][i]==1) 
					{visited[i]=1;
					complex_arrows(x,y,i,visited,complex,lcg);
					visited[i]=0;
					}
			}
		}
}

///////////////////////////////////lcg_check

int lcg_check(int **lcg, int **des, int **complex)

//(Volf and Studeny 1999).

{int i, j, k, l, lcg2, *visited;

visited=new int[n_vars];
for(i=0;i<n_vars;i++)
	{visited[i]=0;
	for(j=0;j<n_vars;j++)
		complex[i][j]=0;
	}

for(i=0;i<n_vars;i++) 
	for(j=0;j<n_vars;j++) 
		if(lcg[i][j]==1 && lcg[j][i]==0 && complex[i][j]==0) 
			{visited[i]=1; 
			visited[j]=1;
			complex_arrows(i,j,j,visited,complex,lcg);
			visited[i]=0; 
			visited[j]=0;
			}

des_build(lcg,des);

lcg2=1;
for(i=0;i<n_vars;i++) 
	for(j=0;j<n_vars;j++)
		if(lcg[i][j]==1 && lcg[j][i]==0 && complex[i][j]==0)
			{lcg2=0;
			for(k=0;k<n_vars;k++) 
				for(l=0;l<n_vars;l++)
					if(complex[k][l]==1 && des[i][k]==1 && des[l][j]==1)
						{lcg2=1;
						k=n_vars;
						l=n_vars;
						}

			if(lcg2==0)
				{i=n_vars;
				j=n_vars;
				}
			}

delete [] visited;

return lcg2;
}

///////////////////////////////////exchange

void exchange(int **x, int **y, int **z, int **w)

{int i, j, aux;

for(i=0;i<n_vars;i++) 
	for(j=0;j<n_vars;j++)
		{aux=x[i][j];
		x[i][j]=y[i][j];
		y[i][j]=aux;
		aux=z[i][j];
		z[i][j]=w[i][j];
		w[i][j]=aux;		
		}
}

///////////////////////////////////sink

int sink(int x, int **cpdag, int *marked)

{int i;

for(i=0;i<n_vars;i++)
	if(marked[i]==0 && cpdag[x][i]==1)
		return 0;

return 1;
}

///////////////////////////////////complete

int complete(int x, int **cpdag, int *marked)

{int i, j;

for(i=0;i<n_vars;i++)
	if(marked[i]==0 && cpdag[i][x]==2)
		for(j=0;j<n_vars;j++)
			if(j!=i && marked[j]==0 && cpdag[j][x]!=0 && cpdag[i][j]==0 && cpdag[j][i]==0)
				return 0;

return 1;
}

///////////////////////////////////cpdag_2_dag 

int cpdag_2_dag(int **cpdag, int **dag)

//(Dor and Tarsi 1992).
//cpdag must be a PDAG according to Dor and Tarsi.
//I think that is not really necessary, i.e. it may contain directed cycles.

{int i, j, k, stop, stop2, *marked;

marked=new int[n_vars];
for(i=0;i<n_vars;i++)
	{marked[i]=0;
	for(j=0;j<n_vars;j++)
		dag[i][j]=0;
	}

stop=n_vars;
while(stop>0)
	{stop2=stop;
	for(j=0;j<n_vars;j++)
		if(marked[j]==0 && sink(j,cpdag,marked)==1 && complete(j,cpdag,marked)==1)
			{stop--;
			marked[j]=1;

			for(k=0;k<n_vars;k++)
				if(marked[k]==0 && cpdag[k][j]!=0)
					dag[k][j]=1; 
			}

	if(stop2==stop)
		{delete [] marked;	
		
		return 0;
		}
	}

delete [] marked;

return 1;
}

///////////////////////////////////mcmc

void mcmc(int &ug, int &dag, double &eq, double &a_l, double &d_l, double &a_a, double &d_a, double &a_c, double &d_c)

//MC starting from the empty LCG.

{int i, j, k, x, y, z, w, *checked, *tocheck, *visited, **lcg, **des, **des2, **complex, **complex2, **closure;

srand((unsigned)time(NULL));

checked=new int[n_vars];
tocheck=new int[n_vars];
visited=new int[n_vars];
lcg=new int*[n_vars];
des=new int*[n_vars];
des2=new int*[n_vars];
complex=new int*[n_vars];
complex2=new int*[n_vars];
closure=new int*[n_vars];
for(i=0;i<n_vars;i++)
	{lcg[i]=new int[n_vars];
	des[i]=new int[n_vars];
	des2[i]=new int[n_vars];
	complex[i]=new int[n_vars];
	complex2[i]=new int[n_vars];
	closure[i]=new int[n_vars];
	for(j=0;j<n_vars;j++)
		lcg[i][j]=0;
	}

lcg_check(lcg,des,complex);

eq=(double)0;
a_l=(double)0;
d_l=(double)0;
a_a=(double)0;
d_a=(double)0;
a_c=(double)0;
d_c=(double)0;
for(k=0;k<n_its;k++)
	{switch(rand()%7)
		{case 0:
			eq++;
			break;
		case 1: //Add line.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && lcg[x][y]==0 && lcg[y][x]==0 && des[x][y]+des[y][x]!=1)
				{lcg[x][y]=1;
				lcg[y][x]=1;
				if(lcg_check(lcg,des2,complex2)==0)
					{lcg[x][y]=0;
					lcg[y][x]=0;
					}
				else
					{exchange(des,des2,complex,complex2);
					a_l++;
					}
				}
			break;
		case 2: //Remove line.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && lcg[x][y]==1 && lcg[y][x]==1)
				{lcg[x][y]=0;
				lcg[y][x]=0;
				if(lcg_check(lcg,des2,complex2)==0)
					{lcg[x][y]=1;
					lcg[y][x]=1;
					}
				else
					{exchange(des,des2,complex,complex2);
					d_l++;
					}
				}
			break;
		case 3: //Add arrow.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && lcg[x][y]==0 && lcg[y][x]==0 && des[y][x]==0)
				{lcg[x][y]=1;
				if(lcg_check(lcg,des2,complex2)==0)
					lcg[x][y]=0;
				else
					{exchange(des,des2,complex,complex2);
					a_a++;
					}
				}
			break;
		case 4: //Remove arrow.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && lcg[x][y]==1 && lcg[y][x]==0)
				{lcg[x][y]=0;
				if(lcg_check(lcg,des2,complex2)==0)
					lcg[x][y]=1;
				else
					{exchange(des,des2,complex,complex2);
					d_a++;
					}
				}
			break;
		case 5: //Add complex.
			x=rand()%n_vars;
			y=rand()%n_vars;
			if(x!=y && lcg[x][y]==0 && lcg[y][x]==0 && des[y][x]==0)
				{lcg[x][y]=1;
				des_build(lcg,des2);
				z=rand()%n_vars;
				w=rand()%n_vars;
				if(x!=z && x!=w && y!=z && z!=w && lcg[z][w]==0 && lcg[w][z]==0 && lcg[x][z]==0 && lcg[z][x]==0 && des2[w][z]==0)
					{lcg[z][w]=1;
					if(lcg_check(lcg,des2,complex2)==0)
						{lcg[x][y]=0;
						lcg[z][w]=0;
						}
					else
						{exchange(des,des2,complex,complex2);
						a_c++;
						}
					}
				else
					lcg[x][y]=0;
				}
			break;
		case 6: //Remove complex.
			x=rand()%n_vars;
			y=rand()%n_vars;
			z=rand()%n_vars;
			w=rand()%n_vars;
			if(x!=y && x!=z && x!=w && y!=z && z!=w && lcg[x][y]==1 && lcg[y][x]==0 && lcg[z][w]==1 && lcg[w][z]==0 && 
			lcg[x][z]==0 && lcg[z][x]==0)
				{lcg[x][y]=0;
				lcg[z][w]=0;
				if(lcg_check(lcg,des2,complex2)==0)
					{lcg[x][y]=1;
					lcg[z][w]=1;
					}
				else
					{exchange(des,des2,complex,complex2);
					d_c++;
					}
				}
			break;
			}
	}

eq/=(double)n_its;
a_l/=(double)n_its;
d_l/=(double)n_its;
a_a/=(double)n_its;
d_a/=(double)n_its;
a_c/=(double)n_its;
d_c/=(double)n_its;

ug=1;
for(i=0;i<n_vars-1;i++)
	for(j=i+1;j<n_vars;j++)
		if(lcg[i][j]+lcg[j][i]==1)
			{ug=0;
			i=n_vars;
			j=n_vars;
			}

//DAG equivalence (Andersson et al. 1997).

for(k=0;k<n_vars;k++)
	checked[k]=0;

dag=1;
for(k=0;k<n_vars;k++)
	if(checked[k]==0)
		{for(i=0;i<n_vars;i++)
			tocheck[i]=0;
			
		for(i=0;i<n_vars;i++)
			if(des[i][k]==1 && des[k][i]==1)
				{checked[i]=1;
				tocheck[i]=1;
				for(j=0;j<n_vars;j++)
					if(lcg[j][i]==1 && lcg[i][j]==0)
						tocheck[j]=1;
				}

		for(i=0;i<n_vars;i++)
			for(j=0;j<n_vars;j++)
				if(tocheck[i]==1 && tocheck[j]==1)
					closure[i][j]=lcg[i][j];
				else
					closure[i][j]=0;

		for(i=0;i<n_vars;i++)
			for(j=0;j<n_vars;j++)
				if(i!=j && tocheck[i]==1 && tocheck[j]==1 && des[k][i]==0 && des[k][j]==0)
					{closure[i][j]=1;
					closure[j][i]=1;
					}

		for(i=0;i<n_vars;i++)
			for(j=0;j<n_vars;j++)
				if(closure[i][j]==1)
					{closure[i][j]=2;
					closure[j][i]=2;
					}

		if(cpdag_2_dag(closure,complex2)==0)
			{dag=0;
			k=n_vars;
			}
		}

delete [] checked;
delete [] tocheck;
delete [] visited;
for(i=0;i<n_vars;i++)
	{delete [] lcg[i];
	delete [] des[i];
	delete [] des2[i];
	delete [] complex[i];
	delete [] complex2[i];
	delete [] closure[i];
	}
delete [] lcg;
delete [] des;
delete [] des2;
delete [] complex;
delete [] complex2;
delete [] closure;
}

///////////////////////////////////main

void main(int argc, char* argv[])

{int i, n_samples, ug, dag, cum_ug, cum_dag, cum_both, cum_none;
double eq, a_l, d_l, a_a, d_a, a_c, d_c, cum_eq, cum_a_l, cum_d_l, cum_a_a, cum_d_a, cum_a_c, cum_d_c;
char name[80];
clock_t clock_init;
FILE *f;

clock_init=clock();

if(argc!=4) 
	{printf_s("\n lcgcounting n_vars n_its n_samples \n"); 
	 exit(0);
	}

n_vars=atoi(argv[1]);
n_its=atoi(argv[2]);
n_samples=atoi(argv[3]);
sprintf_s(name,"lcgcounting_%d.txt",n_vars);

cum_ug=0;
cum_dag=0;
cum_both=0;
cum_none=0;
cum_eq=(double)0;
cum_a_l=(double)0;
cum_d_l=(double)0;
cum_a_a=(double)0;
cum_d_a=(double)0;
cum_a_c=(double)0;
cum_d_c=(double)0;
for(i=0;i<n_samples;i++)
	{mcmc(ug,dag,eq,a_l,d_l,a_a,d_a,a_c,d_c);
	
	cum_ug+=ug;
	cum_dag+=dag;
	if(ug+dag==2)
		cum_both++;
	if(ug+dag==0)
		cum_none++;
	cum_eq+=eq;
	cum_a_l+=a_l;
	cum_d_l+=d_l;
	cum_a_a+=a_a;
	cum_d_a+=d_a;
	cum_a_c+=a_c;
	cum_d_c+=d_c;
	
	if(i%100==99) 
		{printf_s(" (%d, %d, %d, %d) / %d = (%f, %f, %f, %f) in %f hours (%f + %f + %f + %f + %f + %f + %f = %f) \n",cum_ug,cum_dag,
		cum_both,cum_none,i+1,(double)cum_ug/(double)(i+1),(double)cum_dag/(double)(i+1),(double)cum_both/(double)(i+1),
		(double)cum_none/(double)(i+1),(double)(clock()-clock_init)/(double)(CLOCKS_PER_SEC*3600),cum_eq/(double)(i+1),
		cum_a_l/(double)(i+1),cum_d_l/(double)(i+1),cum_a_a/(double)(i+1),cum_d_a/(double)(i+1),cum_a_c/(double)(i+1),
		cum_d_c/(double)(i+1),(cum_eq+cum_a_l+cum_d_l+cum_a_a+cum_d_a+cum_a_c+cum_d_c)/(double)(i+1));
		fopen_s(&f,name,"a");
		fprintf_s(f," (%d, %d, %d, %d) / %d = (%f, %f, %f, %f) in %f hours (%f + %f + %f + %f + %f + %f + %f = %f) \n",cum_ug,cum_dag,
		cum_both,cum_none,i+1,(double)cum_ug/(double)(i+1),(double)cum_dag/(double)(i+1),(double)cum_both/(double)(i+1),
		(double)cum_none/(double)(i+1),(double)(clock()-clock_init)/(double)(CLOCKS_PER_SEC*3600),cum_eq/(double)(i+1),
		cum_a_l/(double)(i+1),cum_d_l/(double)(i+1),cum_a_a/(double)(i+1),cum_d_a/(double)(i+1),cum_a_c/(double)(i+1),
		cum_d_c/(double)(i+1),(cum_eq+cum_a_l+cum_d_l+cum_a_a+cum_d_a+cum_a_c+cum_d_c)/(double)(i+1));
		fclose(f);
		}
	}
}
